# ##############################################################################
#  This file is part of Interdiode                                             #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <matthieu.gallet@19pouces.net>           #
#  All Rights Reserved                                                         #
#                                                                              #
# ##############################################################################
import os
import sys

from django.core.checks import Warning

settings_check_results = []


def missing_package(package_name, desc=""):
    if hasattr(sys, "real_prefix"):  # inside a virtualenv
        cmd = f"Try 'python -m pip install {package_name}' to install it."
    elif __file__.startswith(os.environ.get("HOME", "/home")):
        cmd = f"Try 'python3 -m pip install --user {package_name}' to install it."
    else:
        cmd = f"Try 'sudo python3 -m pip install {package_name}' to install it."
    return Warning(
        f"Python package '{package_name}' is required{desc}. {cmd}",
        obj="configuration",
    )



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <df_config@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
"""Misc functions for the default settings."""
import re
import socket
import sys
from importlib.metadata import PackageNotFoundError, distribution, version
from typing import Dict, Iterable, List, Set
from urllib.parse import urlparse

# noinspection PyPackageRequirements
from django.core.checks import Warning

# noinspection PyPackageRequirements
from django.utils.crypto import get_random_string

from df_config.checks import missing_package, settings_check_results
from df_config.config.dynamic_settings import AutocreateFileContent
from df_config.utils import is_package_present


# noinspection PyUnusedLocal
def get_command_name(settings_dict) -> str:
    """Get the current name."""
    if len(sys.argv) >= 2:
        return sys.argv[1]
    return sys.argv[0] if len(sys.argv) >= 1 else "undefined"


# noinspection PyUnusedLocal
def get_hostname(settings_dict) -> str:
    """Get the current hostname."""
    return socket.gethostname()


def smart_base_url(settings_dict) -> str:
    """Return the external URL of the webserver.

    By default, use the listen address and port as server name.
    Use the "HEROKU_APP_NAME" environment variable if present.

    :param settings_dict:
    :return:
    """
    if settings_dict["HEROKU_APP_NAME"]:
        return "https://%(HEROKU_APP_NAME)s.herokuapp.com/" % settings_dict
    return "http://%(LISTEN_ADDRESS)s/" % settings_dict


smart_base_url.required_settings = ["LISTEN_ADDRESS", "HEROKU_APP_NAME"]


def smart_listen_address(settings_dict):
    """Return the address to listen to with the server command."""
    port = settings_dict["LISTEN_PORT"]
    if isinstance(port, int) and 1 <= port <= 65535:
        return f"0.0.0.0:{port}"
    elif port and re.match(r"^([1-9]\d*)$", port) and 1 <= int(port) <= 65535:
        return f"0.0.0.0:{port}"
    return "localhost:8000"


smart_listen_address.required_settings = ["LISTEN_PORT"]


def template_setting(settings_dict):
    """Return the template settings (taking the DEBUG setting into account)."""
    loaders = [
        "django.template.loaders.filesystem.Loader",
        "django.template.loaders.app_directories.Loader",
    ]
    if settings_dict["DEBUG"]:
        backend = {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
            "NAME": "default",
            "DIRS": settings_dict["TEMPLATE_DIRS"],
            "OPTIONS": {
                "context_processors": settings_dict["TEMPLATE_CONTEXT_PROCESSORS"],
                "loaders": loaders,
                "debug": True,
            },
        }
    else:
        backend = {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
            "NAME": "default",
            "DIRS": settings_dict["TEMPLATE_DIRS"],
            "OPTIONS": {
                "context_processors": settings_dict["TEMPLATE_CONTEXT_PROCESSORS"],
                "debug": False,
                "loaders": [("django.template.loaders.cached.Loader", loaders)],
            },
        }
    return [backend]


template_setting.required_settings = [
    "DEBUG",
    "TEMPLATE_DIRS",
    "TEMPLATE_CONTEXT_PROCESSORS",
]


def allowed_hosts(settings_dict) -> List[str]:
    """Return a list of allow hosts."""
    result = {"127.0.0.1", "::1", "localhost"}
    listened_ip, sep, port = settings_dict["LISTEN_ADDRESS"].rpartition(":")
    if sep == ":" and listened_ip not in ("::", "0.0.0.0"):
        result.add(listened_ip)
    result.add(settings_dict["SERVER_NAME"])
    return list(sorted(result))


allowed_hosts.required_settings = ["SERVER_NAME", "LISTEN_ADDRESS"]


def csrf_trusted_origins(settings_dict) -> List[str]:
    """Return a list of CSRF origins."""
    # noinspection PyPackageRequirements
    from django import VERSION

    if VERSION[0] >= 4:
        # do not append a slash at the end, so cannot reuse SERVER_BASE_URL
        if settings_dict["SERVER_PORT"] == 443 and settings_dict["USE_SSL"]:
            return [
                f"https://{settings_dict['SERVER_NAME']}",
                f"https://{settings_dict['SERVER_NAME']}:443",
            ]
        elif settings_dict["SERVER_PORT"] == 80 and not settings_dict["USE_SSL"]:
            return [
                f"http://{settings_dict['SERVER_NAME']}",
                f"http://{settings_dict['SERVER_NAME']}:80",
            ]
        elif settings_dict["USE_SSL"]:
            return [
                f"https://{settings_dict['SERVER_NAME']}:{settings_dict['SERVER_PORT']}"
            ]
        return [f"http://{settings_dict['SERVER_NAME']}:{settings_dict['SERVER_PORT']}"]
    return [
        f"{settings_dict['SERVER_NAME']}",
        f"{settings_dict['SERVER_NAME']}:{settings_dict['SERVER_PORT']}",
    ]


csrf_trusted_origins.required_settings = ["SERVER_NAME", "SERVER_PORT", "USE_SSL"]


def secure_hsts_seconds(settings_dict) -> int:
    """Return the duration of the HSTS, depending on the use of SSL."""
    if settings_dict["USE_SSL"]:
        return 86400 * 31 * 12
    return 0


secure_hsts_seconds.required_settings = ["USE_SSL"]


def url_parse_server_name(settings_dict) -> str:
    """Return the public hostname, given the public base URL.

    >>> url_parse_server_name({'SERVER_BASE_URL': 'https://demo.example.org/'})
    'demo.example.org'

    """
    return urlparse(settings_dict["SERVER_BASE_URL"]).hostname or "localhost"


url_parse_server_name.required_settings = ["SERVER_BASE_URL"]


def url_parse_server_port(settings_dict) -> int:
    """Return the public port, given the public base URL.

    >>> url_parse_server_port({'SERVER_BASE_URL': 'https://demo.example.org/', 'USE_SSL': True})
    443
    >>> url_parse_server_port({'SERVER_BASE_URL': 'http://demo.example.org/', 'USE_SSL': False})
    80
    >>> url_parse_server_port({'SERVER_BASE_URL': 'https://demo.example.org:8010/', 'USE_SSL': True})
    8010

    """
    port = urlparse(settings_dict["SERVER_BASE_URL"]).port
    https_port = settings_dict["USE_SSL"] and 443
    return port or https_port or 80


url_parse_server_port.required_settings = ["SERVER_BASE_URL", "USE_SSL"]


def url_parse_server_protocol(settings_dict) -> str:
    """Return the public HTTP protocol, given the public base URL.

    >>> url_parse_server_protocol({'USE_SSL': True})
    'https'

    >>> url_parse_server_protocol({'USE_SSL': False})
    'http'

    """
    return "https" if settings_dict["USE_SSL"] else "http"


url_parse_server_protocol.required_settings = ["USE_SSL"]


def url_parse_prefix(settings_dict) -> str:
    """Return the public URL prefix, given the public base URL.

    >>> url_parse_prefix({'SERVER_BASE_URL': 'https://demo.example.org/demo/'})
    '/demo/'
    >>> url_parse_prefix({'SERVER_BASE_URL': 'http://demo.example.org/'})
    '/'
    >>> url_parse_prefix({'SERVER_BASE_URL': 'https://demo.example.org:8010'})
    '/'

    """
    p = urlparse(settings_dict["SERVER_BASE_URL"]).path
    if not p.endswith("/"):
        p += "/"
    return p


url_parse_prefix.required_settings = ["SERVER_BASE_URL"]


def url_parse_ssl(settings_dict) -> bool:
    """Return True if the public URL uses HTTPS.

    >>> url_parse_ssl({'SERVER_BASE_URL': 'https://demo.example.org/demo/'})
    True
    >>> url_parse_ssl({'SERVER_BASE_URL': 'http://demo.example.org/'})
    False

    """
    return urlparse(settings_dict["SERVER_BASE_URL"]).scheme == "https"


url_parse_ssl.required_settings = ["SERVER_BASE_URL"]


def use_x_forwarded_for(settings_dict) -> bool:
    """Return `True` if this server is assumed to be behind a reverse proxy.

     Heuristic: the external port (in SERVER_PORT) is different from the actually listened port (in LISTEN_ADDRESS).

    >>> use_x_forwarded_for({'SERVER_PORT': 8000, 'LISTEN_ADDRESS': 'localhost:8000'})
    False
    >>> use_x_forwarded_for({'SERVER_PORT': 443, 'LISTEN_ADDRESS': 'localhost:8000'})
    True

    """
    listen_address, sep, listen_port = settings_dict["LISTEN_ADDRESS"].rpartition(":")
    if not re.match(r"\d+", listen_port):
        raise ValueError("Invalid LISTEN_ADDRESS port %s" % listen_port)
    return int(listen_port) != settings_dict["SERVER_PORT"]


use_x_forwarded_for.required_settings = ["SERVER_PORT", "LISTEN_ADDRESS"]


def project_name(settings_dict) -> str:
    """Transform the base module name into a nicer project name.

    >>> project_name({'DF_MODULE_NAME': 'my_project'})
    'My Project'
    """
    return " ".join(
        [
            x.capitalize()
            for x in settings_dict["DF_MODULE_NAME"].replace("_", " ").split()
        ]
    )


project_name.required_settings = ["DF_MODULE_NAME"]


class AutocreateSecretKey(AutocreateFileContent):
    """Generate a secret key and store it in a file."""

    def __init__(self, filename):
        """Generate a secret key and store it in a file."""
        super().__init__(filename, generate_secret_key, mode=0o600, length=60)


def generate_secret_key(django_ready, length=60) -> str:
    """Generate a default random secret key."""
    if not django_ready:
        return get_random_string(length=length)
    # noinspection PyPackageRequirements
    from django.conf import settings

    return settings.SECRET_KEY


def required_packages(settings_dict) -> List[str]:
    """Return a sorted list of the Python packages required by the current project.

    Issue  a warning for each missing package.
    """
    checked_packages: Set[str] = set()

    def get_requirements(package_name, parent=None) -> Iterable[str]:
        if package_name not in checked_packages:
            checked_packages.add(package_name)
            try:
                yield str(package_name)
                d = distribution(package_name)
                for r in d.requires:
                    r, __, __ = r.partition(";")
                    r, __, __ = r.partition("(")
                    r = r.strip()
                    for required_package in get_requirements(r, parent=package_name):
                        yield str(required_package)
            except PackageNotFoundError:
                settings_check_results.append(
                    missing_package(str(package_name), f" by {parent}")
                )
            except Exception as e:
                settings_check_results.append(
                    missing_package(str(package_name), f" by {parent} ({e})")
                )

    return list(
        sorted(
            set(
                get_requirements(
                    settings_dict["DF_MODULE_NAME"],
                    parent=settings_dict["DF_MODULE_NAME"],
                )
            )
        )
    )


required_packages.required_settings = ["DF_MODULE_NAME"]


class ExcludedDjangoCommands:
    """Exclude some Django commands that have no use of the end-user."""

    required_settings = ["DEVELOPMENT", "USE_CELERY", "DEBUG"]

    def __call__(self, settings_dict):
        """Exclude django commands in production mode."""
        result = {"startproject", "diffsettings"}
        if not settings_dict["DEVELOPMENT"]:
            result |= {
                "startapp",
                "findstatic",
                "npm",
                "makemigrations",
                "makemessages",
                "inspectdb",
                "compilemessages",
                "remove_stale_contenttypes",
                "squashmigrations",
            }
        if not settings_dict["USE_CELERY"]:
            result |= {"celery", "worker"}
        if not settings_dict["DEBUG"] and not settings_dict["DEVELOPMENT"]:
            result |= {"testserver", "test", "runserver"}
        return result

    def __repr__(self):
        """Return a useful representation."""
        return "%s.%s" % (self.__module__, "excluded_django_commands")


excluded_django_commands = ExcludedDjangoCommands()


def get_asgi_application(settings_dict) -> str:
    """Return the ASGI application (depends of the use of websockets)."""
    if settings_dict["USE_WEBSOCKETS"]:
        application = "df_websockets.routing.application"
    else:
        application = "df_config.application.asgi_application"
    return application


get_asgi_application.required_settings = ["USE_WEBSOCKETS"]


# noinspection PyUnusedLocal
def get_wsgi_application(settings_dict) -> str:
    """Return the WSGI application."""
    return "df_config.application.wsgi_application"


get_wsgi_application.required_settings = []


def use_sentry(settings_dict: Dict) -> bool:
    """Guess if Sentry is available."""
    sentry_dsn = settings_dict["SENTRY_DSN"]
    if not sentry_dsn:
        return False
    if not is_package_present("sentry_sdk"):
        settings_check_results.append(
            Warning("sentry_sdk must be installed.", obj="configuration")
        )
        return False
    # noinspection PyUnresolvedReferences
    import sentry_sdk

    # noinspection PyUnresolvedReferences
    from sentry_sdk.integrations.django import DjangoIntegration

    integrations = [DjangoIntegration()]
    if settings_dict["USE_CELERY"]:
        # noinspection PyUnresolvedReferences
        from sentry_sdk.integrations.celery import CeleryIntegration

        integrations.append(CeleryIntegration())
    sentry_sdk.init(
        dsn=sentry_dsn,
        integrations=integrations,
        traces_sample_rate=1.0,
        debug=settings_dict["DEBUG"],
        send_default_pii=True,
    )
    return True


use_sentry.required_settings = ["SENTRY_DSN", "USE_CELERY", "DEBUG"]


# noinspection PyUnusedLocal
def web_server(settings_dict) -> str:
    """Provide a valid server (daphne if available, uvicorn or gunicorn)."""
    for server_name in "daphne", "gunicorn", "uvicorn":
        if is_package_present(server_name):
            return server_name
    return "gunicorn"


web_server.required_settings = []


def csp_connect(settings_dict) -> List[str]:
    """Return a valid value for the CSP connect option."""
    values = ["'self'"]
    if settings_dict.get("USE_SSL") and settings_dict.get("USE_WEBSOCKETS"):
        values.append("wss://%(SERVER_NAME)s:%(SERVER_PORT)s" % settings_dict)
    elif settings_dict.get("USE_WEBSOCKETS"):
        values.append("ws://%(SERVER_NAME)s:%(SERVER_PORT)s" % settings_dict)
    return values


web_server.required_settings = [
    "USE_SSL",
    "SERVER_NAME",
    "USE_WEBSOCKETS",
    "SERVER_PORT",
]



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <df_config@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
import string
import sys
from collections import OrderedDict

from django.core.management import color_style
from django.core.management.base import OutputWrapper
from django.core.management.color import no_style

from df_config.config.dynamic_settings import DynamicSettting, ExpandIterable
from df_config.config.fields import ConfigField
from df_config.config.fields_providers import PythonConfigFieldsProvider
from df_config.config.values_providers import ConfigProvider


class SettingMerger:
    """Load different settings modules and config files and merge them."""

    def __init__(
        self,
        fields_provider,
        providers,
        stdout=None,
        stderr=None,
        no_color=False,
    ):
        self.fields_provider = fields_provider or PythonConfigFieldsProvider(None)
        self.providers = providers or []
        self.__formatter = string.Formatter()
        self.settings = {}
        self.config_values = (
            []
        )  # list of (ConfigValue, provider_name, setting_name, final_value)
        self.raw_settings = OrderedDict()
        # raw_settings[setting_name][str(provider) or None] = raw_value
        self.__working_stack = set()
        self.stdout = OutputWrapper(stdout or sys.stdout)
        self.stderr = OutputWrapper(stderr or sys.stderr)
        if no_color:
            self.style = no_style()
        else:
            self.style = color_style()
            self.stderr.style_func = self.style.ERROR

    def add_provider(self, provider):
        self.providers.append(provider)

    def process(self):
        self.load_raw_settings()
        self.load_settings()

    def load_raw_settings(self):
        # get all setting names and sort them
        all_settings_names_set = set()
        for field in self.fields_provider.get_config_fields():
            assert isinstance(field, ConfigField)
            all_settings_names_set.add(field.setting_name)
        for provider in self.providers:
            assert isinstance(provider, ConfigProvider)
            for setting_name, value in provider.get_extra_settings():
                all_settings_names_set.add(setting_name)
        all_settings_names = list(sorted(all_settings_names_set))
        # initialize all defined settings
        for setting_name in all_settings_names:
            self.raw_settings[setting_name] = OrderedDict()
        # fetch default values if its exists (useless?)
        for field in self.fields_provider.get_config_fields():
            assert isinstance(field, ConfigField)
            self.raw_settings[field.setting_name][None] = field.value
        # read all providers (in the right order)
        for provider in self.providers:
            assert isinstance(provider, ConfigProvider)
            source_name = str(provider)
            for field in self.fields_provider.get_config_fields():
                assert isinstance(field, ConfigField)
                if provider.has_value(field):
                    value = provider.get_value(field)
                    # noinspection PyTypeChecker
                    self.raw_settings[field.setting_name][source_name] = value
            for setting_name, value in provider.get_extra_settings():
                self.raw_settings[setting_name][source_name] = value

    def has_setting_value(self, setting_name):
        return setting_name in self.raw_settings

    def get_setting_value(self, setting_name):
        if setting_name in self.settings:
            return self.settings[setting_name]
        elif setting_name in self.__working_stack:
            raise ValueError(
                "Invalid cyclic dependency between " + ", ".join(self.__working_stack)
            )
        elif setting_name not in self.raw_settings:
            raise ValueError("Invalid setting reference: %s" % setting_name)
        self.__working_stack.add(setting_name)
        provider_name, raw_value = None, None
        for provider_name, raw_value in self.raw_settings[setting_name].items():
            pass
        value = self.analyze_raw_value(raw_value, provider_name, setting_name)
        self.settings[setting_name] = value
        self.__working_stack.remove(setting_name)
        return value

    def load_settings(self):
        for setting_name in self.raw_settings:
            self.get_setting_value(setting_name)

    def call_method_on_config_values(self, method_name: str):
        """Scan all settings, looking for :class:`df_config.config.dynamic_settings.DynamicSetting` and calling one of their
        methods.

        :param method_name: 'pre_collectstatic', 'pre_migrate', 'post_collectstatic', or 'post_migrate'.
        """
        for raw_value, provider_name, setting_name, final_value in self.config_values:
            try:
                getattr(raw_value, method_name)(
                    self, provider_name, setting_name, final_value
                )
            except Exception as e:
                self.stdout.write(
                    self.style.ERROR(
                        'Invalid value "%s" in %s for %s (%s)'
                        % (raw_value, provider_name or "built-in", setting_name, e)
                    )
                )

    def analyze_raw_value(self, obj, provider_name, setting_name):
        """Parse the object for replacing variables by their values.

        If `obj` is a string like "THIS_IS_{TEXT}", search for a setting named "TEXT" and replace {TEXT} by its value
        (say, "VALUE"). The returned object is then equal to "THIS_IS_VALUE".
        If `obj` is a list, a set, a tuple or a dict, its components are recursively parsed.
        If `obj` is a subclass of :class:`df_config.config.dynamic_settings.DynamicSetting`, its value is on-the-fly computed.
        Otherwise, `obj` is returned as-is.

        :param obj: object to analyze
        :param provider_name: the name of the config file
        :param setting_name: the name of the setting containing this value
            but this value can be inside a dict or a list (like `SETTING = [Directory("/tmp"), ]`)
        :return: the parsed setting
        """
        if isinstance(obj, str):
            values = {}
            for (
                literal_text,
                field_name,
                format_spec,
                conversion,
            ) in self.__formatter.parse(obj):
                if field_name is not None:
                    values[field_name] = self.get_setting_value(field_name)
            return self.__formatter.format(obj, **values)
        elif isinstance(obj, DynamicSettting):
            final_value = obj.get_value(self, provider_name, setting_name)
            self.config_values.append((obj, provider_name, setting_name, final_value))
            return final_value
        elif isinstance(obj, list) or isinstance(obj, tuple):
            result = []
            for sub_obj in obj:
                if isinstance(sub_obj, ExpandIterable):
                    result += self.get_setting_value(sub_obj.value)
                else:
                    result.append(
                        self.analyze_raw_value(sub_obj, provider_name, setting_name)
                    )
            if isinstance(obj, tuple):
                return tuple(result)
            return result
        elif isinstance(obj, set):
            result = set()
            for sub_obj in obj:
                if isinstance(sub_obj, ExpandIterable):
                    result |= self.get_setting_value(sub_obj.value)
                else:
                    result.add(
                        self.analyze_raw_value(sub_obj, provider_name, setting_name)
                    )
            return result
        elif isinstance(obj, dict):
            result = {}  # OrderedDict or plain dict
            for sub_key, sub_obj in obj.items():
                if isinstance(sub_obj, ExpandIterable):
                    result.update(self.get_setting_value(sub_obj.value))
                else:
                    value = self.analyze_raw_value(sub_obj, provider_name, setting_name)
                    key = self.analyze_raw_value(sub_key, provider_name, setting_name)
                    result[key] = value
            # to work with OrderedDict and defaultdict, we need to use the provided object and not a new one
            to_remove = [key for key in obj if key not in result]
            for key in to_remove:
                del obj[key]
            obj.update(result)
            return obj
        return obj

    def post_process(self):
        """Perform some cleaning on settings:

        * remove duplicates in `INSTALLED_APPS` (keeps only the first occurrence)
        """
        # remove duplicates in INSTALLED_APPS
        key = "INSTALLED_APPS"
        if key in self.settings:
            self.settings[key] = list(OrderedDict.fromkeys(self.settings[key]))

    def write_provider(self, provider, include_doc=False):
        config_fields = self.fields_provider.get_config_fields()
        for config_field in sorted(config_fields, key=lambda x: str(x.name)):
            assert isinstance(config_field, ConfigField)
            if config_field.setting_name not in self.settings:
                continue
            config_field.value = self.settings[config_field.setting_name]
            provider.set_value(config_field, include_doc=include_doc)



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <df_config@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
"""Provide a meaningful log configuration, depending on several options."""
import logging
import logging.handlers
import os
import re
import sys
import time
import warnings
from traceback import extract_stack
from urllib.parse import urlparse

from django.core.checks.messages import Warning
from django.core.management import color_style
from django.utils.log import AdminEmailHandler as BaseAdminEmailHandler

from df_config.checks import settings_check_results


class ColorizedFormatter(logging.Formatter):
    """Used in console for applying colors to log lines, corresponding to the log level."""

    def __init__(self, *args, **kwargs):
        """Initialize the formatter."""
        self.style = color_style()
        kwargs.setdefault("fmt", "%(asctime)s [%(name)s] [%(levelname)s] %(message)s")
        kwargs.setdefault("datefmt", "%Y-%m-%d %H:%M:%S")
        super().__init__(*args, **kwargs)

    def format(self, record):
        """Apply a log color, corresponding to the log level."""
        msg = record.msg
        level = record.levelno
        if level <= logging.DEBUG:
            msg = self.style.HTTP_SUCCESS(msg)
        elif level <= logging.INFO:
            msg = self.style.HTTP_NOT_MODIFIED(msg)
        elif level <= logging.WARNING:
            msg = self.style.WARNING(msg)
        else:
            msg = self.style.ERROR(msg)
        record.msg = msg
        return super().format(record)

    def formatStack(self, stack_info):
        """Colorize errors in red."""
        return self.style.ERROR(stack_info)


class ServerFormatter(logging.Formatter):
    """Formatter for the access logs."""

    def __init__(self, *args, **kwargs):
        """Initialize the object."""
        self.style = color_style()
        super().__init__(*args, **kwargs)

    def format(self, record):
        """Format an access, colorizing it depending on the status code."""
        msg = record.msg
        status_code = getattr(record, "status_code", None)
        level = record.levelno
        if status_code:
            if 200 <= status_code < 300:
                # Put 2XX first, since it should be the common case
                msg = self.style.SUCCESS(msg)
            elif 100 <= status_code < 200:
                msg = self.style.HTTP_INFO(msg)
            elif status_code == 304:
                msg = self.style.HTTP_NOT_MODIFIED(msg)
            elif 300 <= status_code < 400:
                msg = self.style.HTTP_REDIRECT(msg)
            elif status_code == 404:
                msg = self.style.HTTP_NOT_FOUND(msg)
            elif 400 <= status_code < 500:
                msg = self.style.HTTP_BAD_REQUEST(msg)
            else:
                # Any 5XX, or any other status code
                msg = self.style.HTTP_SERVER_ERROR(msg)
        elif level <= logging.DEBUG:
            msg = self.style.HTTP_SUCCESS(msg)
        elif level <= logging.INFO:
            msg = self.style.HTTP_NOT_MODIFIED(msg)
        elif level <= logging.WARNING:
            msg = self.style.WARNING(msg)
        else:
            msg = self.style.ERROR(msg)

        if self.uses_server_time() and not hasattr(record, "server_time"):
            record.server_time = self.formatTime(record, self.datefmt)

        record.msg = msg
        return super().format(record)

    def uses_server_time(self):
        """Return true if the log format requires the response time."""
        return self._fmt.find("%(server_time)") >= 0


# noinspection PyClassHasNoInit
class AdminEmailHandler(BaseAdminEmailHandler):
    """Enhance the AdminEmailHandler provided by Django.

    Does not try to send emails if `settings.EMAIL_HOST` is not set.
    Also limits the mail rates to avoid to spam the poor admins.
    """

    _previous_email_time = None
    min_interval = 600
    """min time (in seconds) between two successive sends"""

    def send_mail(self, subject, message, *args, **kwargs):
        """Check if email can be sent before applying the original method."""
        # noinspection PyPackageRequirements
        from django.conf import settings

        if self.can_send_email() and settings.EMAIL_HOST:
            try:
                super().send_mail(subject, message, *args, **kwargs)
            except Exception as e:
                print(
                    "Unable to send e-mail to admin. Please checks your e-mail settings [%r]."
                    % e
                )
                if settings.LOG_DIRECTORY:
                    print("Check logs in %s" % settings.LOG_DIRECTORY)

    def can_send_email(self):
        """Check the time of the previous email to allow the new one."""
        now = time.time()
        previous = AdminEmailHandler._previous_email_time
        AdminEmailHandler._previous_email_time = now
        can_send = True
        if previous and now - previous < self.min_interval:
            can_send = False
        return can_send


class RemoveDuplicateWarnings(logging.Filter):
    """Displays py.warnings messages unless the same warning was already sent."""

    def __init__(self, name=""):
        """Init function."""
        super().__init__(name=name)
        self.previous_records = set()

    def filter(self, record):
        """Check if the message has already been sent from the same Python file."""
        record_value = hash("%r %r" % (record.pathname, record.args))
        result = record_value not in self.previous_records
        self.previous_records.add(record_value)
        return result


class SlowQueriesFilter(logging.Filter):
    """Filter slow queries and attach stack_info."""

    def __init__(self, name="", slow_query_duration_in_s=1.0):
        """Init function."""
        super().__init__(name=name)
        self.slow_query_duration_in_s = slow_query_duration_in_s

    def filter(self, record):
        """Filter SQL queries depending on their duration."""
        duration = getattr(record, "duration", 0)
        if duration > self.slow_query_duration_in_s:
            # Same as in _log for when stack_info=True is used.
            # noinspection PyTypeChecker
            fn, lno, func, sinfo = logging.Logger.findCaller(None, True)
            record.stack_info = sinfo
            return True
        return False


# noinspection PyMethodMayBeStatic
class LogConfiguration:
    """Generate a log configuration depending on a few parameters.

    * the debug mode (if `DEBUG == True`, everything is printed to the console and lower log level are applied),
    * the log directory (if set, everything is output to several rotated log files),
    * the log remote URL (to send data to syslog or logd),
    * script name (for determining the log filename).

    Required values in the `settings_dict`:

    * `DEBUG`: `True` or `False`
    * `DF_MODULE_NAME`: your project name, used to determine log filenames,
    * `LOG_DIRECTORY`: dirname where log files are written (!),
    * `LOG_LEVEL`: one of "debug", "info", "warning", "error", "critical"
    * `LOG_REMOTE_URL`: examples: "syslog+tcp://localhost:514/user", "syslog:///local7"
         "syslog:///dev/log/daemon", "logd:///project_name"
    * `LOG_REMOTE_ACCESS`: also send HTTP requests to syslog/journald
    * `LOG_SLOW_QUERY_DURATION_IN_S`: log requests that takes more than this time
    * `SERVER_NAME`: the public name of the server (like "www.example.com")
    * `SERVER_PORT`: the public port (probably 80 or 443)
    * `LOG_EXCLUDED_COMMANDS`: Django commands that do not write logs
    """

    required_settings = [
        "DEBUG",
        "DF_MODULE_NAME",
        "LOG_DIRECTORY",
        "LOG_REMOTE_URL",
        "LOG_SLOW_QUERY_DURATION_IN_S",
        "LOG_REMOTE_ACCESS",
        "SERVER_NAME",
        "SERVER_PORT",
        "LOG_EXCLUDED_COMMANDS",
        "LOG_LEVEL",
    ]
    # for loggers that only show INFO in debug mode, or WARNING in INFO, and so on:
    _level_up = {
        "DEBUG": "INFO",
        "INFO": "WARNING",
        "WARNING": "ERROR",
        "ERROR": "CRITICAL",
    }
    # set of always INFO loggers that are not propagated to root
    _level_access = {
        "DEBUG": "INFO",
        "WARNING": "INFO",
        "ERROR": "INFO",
        "CRITICAL": "INFO",
    }
    access_loggers = {
        "aiohttp.access": _level_access,
        "django.server": _level_access,
        "django.channels.server": _level_access,
        "geventwebsocket.handler": _level_access,
        "gunicorn.access": _level_access,
        "uvicorn.access": _level_access,
    }

    problem_loggers = {
        "django": _level_up,
        "django.db": _level_up,
        "django.db.backends.schema": _level_up,
        "django.db.backends": {
            "DEBUG": "DEBUG",
            "INFO": "DEBUG",
            "WARNING": "DEBUG",
            "ERROR": "ERROR",
            "CRITICAL": "CRITICAL",
        },
        "django.request": {},
        "django.security": {},
        "df_websockets.signals": {},
        "gunicorn.error": {},
        "uvicorn.error": {},
        "pip.vcs": _level_up,
        "py.warnings": _level_up,
    }
    compat_log_levels = {
        "WARN": "WARNING",
        "CRIT": "CRITICAL",
        "EMERGENCY": "CRITICAL",
        "ALERT": "CRITICAL",
        "NOTICE": "INFO",
    }

    # all loggers that will be defined
    # values are dict to map chosen log level (by the admin) to the log level

    def __init__(self, stdout=None, stderr=None):
        """Init function."""
        self.formatters = {}
        self.filters = {}
        self.loggers = {}
        self.handlers = {}
        self.root = {}
        self.log_suffix = None
        self.module_name = None
        self.log_directory = None
        self.server_name = None
        self.log_level = None
        self.server_port = None
        self.slow_query_duration_in_s = None
        self.excluded_commands = {}
        self.stdout = stdout or sys.stdout
        self.stderr = stderr or sys.stderr
        self.log_directory_warning = False  # True when a warning has been emitted

    def __call__(self, settings_dict, argv=None):
        """Create the log configuration during the setting computation."""
        if argv is None:
            argv = sys.argv
        self.module_name = settings_dict["DF_MODULE_NAME"]
        self.server_name = settings_dict["SERVER_NAME"]
        self.server_port = settings_dict["SERVER_PORT"]
        self.slow_query_duration_in_s = settings_dict["LOG_SLOW_QUERY_DURATION_IN_S"]
        self.excluded_commands = settings_dict["LOG_EXCLUDED_COMMANDS"]
        if settings_dict["LOG_LEVEL"]:
            log_level = settings_dict["LOG_LEVEL"].upper()
        elif settings_dict["DEBUG"]:
            log_level = "DEBUG"
        else:
            log_level = "WARNING"
        log_level = self.compat_log_levels.get(log_level, log_level)
        self.formatters = self.get_default_formatters()
        self.filters = self.get_default_filters()
        self.loggers = self.get_default_loggers()
        self.handlers = self.get_default_handlers()
        self.root = self.get_default_root()
        self.log_suffix = self.get_smart_command_name(
            self.module_name,
            argv,
            self.excluded_commands,
        )
        self.log_directory = settings_dict["LOG_DIRECTORY"]
        config = {
            "version": 1,
            "disable_existing_loggers": True,
            "formatters": self.formatters,
            "filters": self.filters,
            "handlers": self.handlers,
            "loggers": self.loggers,
            "root": self.root,
        }
        self.root["level"] = log_level
        for logger, levels in self.problem_loggers.items():
            self.loggers[logger]["level"] = levels.get(log_level, log_level)
        for logger, levels in self.access_loggers.items():
            self.loggers[logger]["level"] = levels.get(log_level, log_level)
        if settings_dict["DEBUG"]:
            warnings.simplefilter("always", DeprecationWarning)
            logging.captureWarnings(True)
            self.add_handler("ROOT", "stdout", level="INFO", formatter="colorized")
            for logger in self.access_loggers:
                self.add_handler(
                    logger, "stderr", level="DEBUG", formatter="django.server"
                )
            return config

        has_handler = False

        if self.log_directory and self.log_suffix:
            self.add_handler("ROOT", "root", level=log_level)
            for logger in self.access_loggers:
                self.add_handler(logger, "access", level="DEBUG", formatter="nocolor")
            has_handler = True

        has_handler = (
            self.add_remote_collector(
                settings_dict["LOG_REMOTE_URL"],
                settings_dict["LOG_REMOTE_ACCESS"],
                level=log_level,
            )
            or has_handler
        )
        if not has_handler or not self.log_suffix:
            # (no file or interactive command) and no logd/syslog => we print to the console (like the debug mode)
            self.add_handler("ROOT", "stdout", level=log_level, formatter="colorized")
            for logger in self.access_loggers:
                self.add_handler(
                    logger, "stderr", formatter="django.server", level=log_level
                )
        self.root["handlers"].append("mail_admins")
        return config

    def __repr__(self):
        """Return a valid representation."""
        return "%s.%s" % (self.__module__, "log_configuration")

    def add_remote_collector(self, log_remote_url, log_remote_access, level="WARNING"):
        """Add a remote collector, like syslog or loki."""
        has_handler = False
        if not log_remote_url:
            return has_handler
        parsed_log_url = urlparse(log_remote_url)
        scheme = parsed_log_url.scheme
        device, sep, facility_name = parsed_log_url.path.rpartition("/")
        if scheme == "syslog" or scheme == "syslog+tcp":
            address, facility, socktype = self.parse_syslog_url(
                parsed_log_url, scheme, device, facility_name
            )
            kwargs = {
                "address": address,
                "facility": facility,
                "socktype": socktype,
                "formatter": "nocolor",
            }
            self.add_handler("ROOT", "syslog", level=level, **kwargs)
            if log_remote_access:
                for logger in self.access_loggers:
                    self.add_handler(logger, "syslog", level="DEBUG", **kwargs)
            has_handler = True
        elif scheme == "loki" or scheme == "lokis":
            url = f"http://{parsed_log_url.hostname}"
            if scheme == "lokis":
                url = f"https://{parsed_log_url.hostname}"
            if parsed_log_url.port:
                url += f":{parsed_log_url.port}"
            if parsed_log_url.path:
                url += parsed_log_url.path
            if parsed_log_url.query:
                url += f"?{parsed_log_url}"
            auth = None
            if parsed_log_url.username and parsed_log_url.password:
                auth = (parsed_log_url.username, parsed_log_url.password)
            kwargs = {"url": url, "auth": auth}
            self.add_handler("ROOT", "loki", level=level, **kwargs)
            if log_remote_access:
                for logger in self.access_loggers:
                    self.add_handler(logger, "loki", level="DEBUG", **kwargs)
            has_handler = True
        else:
            warning = Warning(
                "The only known schemes for remote logging are syslog, syslog+tcp, loki or lokis.",
                hint=None,
                obj="configuration",
            )
            settings_check_results.append(warning)
        return has_handler

    def parse_syslog_url(self, parsed_log_url, scheme, device, facility_name):
        """Parse a syslog URL and return valid parameters."""
        import platform
        import socket
        import syslog

        if (
            parsed_log_url.hostname
            and parsed_log_url.port
            and re.match(r"^\d+$", str(parsed_log_url.port))
        ):
            address = (parsed_log_url.hostname, int(parsed_log_url.port))
        elif device:
            address = device
        elif platform.system() == "Darwin":
            address = "/var/run/syslog"
        elif platform.system() == "Linux":
            address = "/dev/log"
        else:
            address = ("localhost", 514)
        socktype = socket.SOCK_DGRAM if scheme == "syslog" else socket.SOCK_STREAM
        # noinspection PyUnresolvedReferences
        facility = logging.handlers.SysLogHandler.facility_names.get(
            facility_name, syslog.LOG_USER
        )
        return address, facility, socktype

    @property
    def fmt_stderr(self):
        """Return the valid formatter for stderr (if it's a TTY)."""
        return "colorized" if self.stderr.isatty() else None

    @property
    def fmt_stdout(self):
        """Return the valid formatter for stderr (if it's a TTY)."""
        return "colorized" if self.stdout.isatty() else None

    def get_default_formatters(self):
        """Return some default formatters."""
        name = "%s:%s" % (self.server_name, self.server_port)
        return {
            "django.server": {
                "()": "df_config.guesses.log.ServerFormatter",
                "format": "%(asctime)s [{}] %(message)s".format(name),
            },
            "nocolor": {
                "()": "logging.Formatter",
                "fmt": "%(asctime)s [{}] [%(levelname)s] %(message)s".format(name),
                "datefmt": "%Y-%m-%d %H:%M:%S",
            },
            "colorized": {"()": "df_config.guesses.log.ColorizedFormatter"},
        }

    def get_default_handlers(self):
        """Return default handlers."""
        return {
            "mail_admins": {
                "class": "df_config.guesses.log.AdminEmailHandler",
                "level": "ERROR",
                "include_html": True,
            }
        }

    def get_default_filters(self):
        """Return default filters."""
        filters = {
            "remove_duplicate_warnings": {
                "()": "df_config.guesses.log.RemoveDuplicateWarnings"
            },
        }
        if self.slow_query_duration_in_s:
            filters["slow_queries"] = {
                "()": "df_config.guesses.log.SlowQueriesFilter",
                "slow_query_duration_in_s": self.slow_query_duration_in_s,
            }

        return filters

    def get_default_root(self):
        """Return the default log root."""
        return {"handlers": [], "level": "WARNING"}

    def get_default_loggers(self):
        """Return the default loggers."""
        loggers = {}
        for logger in self.problem_loggers:
            loggers[logger] = {"handlers": [], "level": "DEBUG", "propagate": True}
        if self.slow_query_duration_in_s:
            loggers["django.db.backends"]["filters"] = ["slow_queries"]
        loggers["py.warnings"]["filters"] = ["remove_duplicate_warnings"]
        for logger in self.access_loggers:
            loggers[logger] = {"handlers": [], "level": "DEBUG", "propagate": False}
        return loggers

    def add_handler(
        self,
        logger: str,
        filename: str,
        level: str = "WARNING",
        formatter=None,
        **kwargs,
    ):
        """Add a handler to a logger.

        The name of the added handler is unique, so the definition of the handler is also add if required.
        You can use "ROOT" as logger name to target the root logger.

        filename: can be a filename or one of the following special values: "stderr", "stdout", "logd", "syslog"
        """
        if filename == "stderr":
            handler, handler_name = self.add_handler_stderr(filename, formatter, level)
        elif filename == "stdout":
            handler, handler_name = self.add_handler_stdout(filename, formatter, level)
        elif filename == "loki":
            handler, handler_name = self.add_handler_loki(
                logger, filename, level, kwargs
            )
        elif filename == "syslog":
            handler_name = f"{filename}.{level.lower()}"
            handler = {"class": "logging.handlers.SysLogHandler", "level": level}
            handler.update(kwargs)
        elif filename == "logd":
            handler, handler_name = self.add_handler_logd(
                logger, filename, level, kwargs
            )
        elif self.log_directory:  # basename of a plain-text log
            handler, handler_name = self.add_handler_directory(
                logger, filename, level, kwargs
            )
        else:
            handler, handler_name = None, None
        if not handler_name:
            return
        if handler_name not in self.handlers:
            self.handlers[handler_name] = handler
        if logger == "ROOT":
            target = self.root
        else:
            target = self.loggers[logger]
        if handler_name not in target["handlers"]:
            target["handlers"].append(handler_name)

    def add_handler_loki(self, logger, filename, level, kwargs):
        """Add a loki handler when required and possible."""
        try:
            # noinspection PyUnresolvedReferences,PyPackageRequirements
            import logging_loki
        except ImportError:
            warning = Warning(
                "Unable to import logging_loki (required to log to Loki)",
                hint=None,
                obj="configuration",
            )
            settings_check_results.append(warning)
            # replace loki by writing to a plain-text log
            self.add_handler(logger, level.lower(), level=level)
            return None, None
        handler_name = f"{filename}.{level.lower()}"
        handler = {"class": "df_config.extra.loki.LokiHandler", "level": level}
        handler.update(kwargs)
        return handler, handler_name

    def add_handler_logd(self, logger, filename, level, kwargs):
        """Add a logd (systemd) handler when required and possible."""
        try:
            # noinspection PyUnresolvedReferences,PyPackageRequirements
            import systemd.journal
        except ImportError:
            warning = Warning(
                "Unable to import systemd.journal (required to log with journlad)",
                hint=None,
                obj="configuration",
            )
            settings_check_results.append(warning)
            # replace logd by writing to a plain-text log
            self.add_handler(logger, level.lower(), level=level)
            return None, None
        handler_name = f"{filename}.{level.lower()}"
        handler = {"class": "systemd.journal.JournalHandler", "level": level}
        handler.update(kwargs)
        return handler, handler_name

    def add_handler_directory(self, logger, filename, level, kwargs):
        """Add a log handler when a log directory is defined and writeable."""
        log_directory = os.path.normpath(self.log_directory)
        if not os.path.isdir(log_directory):
            if not self.log_directory_warning:
                warning = Warning(
                    f'Missing directory "{log_directory}"',
                    hint=None,
                    obj=log_directory,
                )
                settings_check_results.append(warning)
                self.log_directory_warning = True
            self.add_handler(logger, "stdout", level=level, **kwargs)
            return None, None
        basename = f"{self.log_suffix}-{filename}.log"
        log_filename = os.path.join(log_directory, basename)
        try:
            remove = not os.path.exists(log_filename)
            open(log_filename, "a").close()  # ok, we can write
            if (
                remove
            ):  # but if this file did not exist, we remove it to avoid lot of empty log files...
                os.remove(log_filename)
        except PermissionError:
            warning_ = Warning(
                f'Unable to write logs in "{log_directory}". Unsufficient rights?',
                hint=None,
                obj=log_directory,
            )
            settings_check_results.append(warning_)
            self.add_handler(logger, "stdout", level=level, **kwargs)
            return None, None
        handler_name = "%s.%s" % (self.log_suffix, filename)
        handler = {
            "class": "logging.handlers.RotatingFileHandler",
            "maxBytes": 1000000,
            "backupCount": 3,
            "formatter": "nocolor",
            "filename": log_filename,
            "level": level,
            "delay": True,
        }
        return handler, handler_name

    def add_handler_stdout(self, filename, formatter, level):
        """Add an handler for stdout."""
        handler_name = f"{filename}.{level.lower()}"
        if formatter in ("django.server", "colorized") and not self.stdout.isatty():
            formatter = None
        elif formatter:
            handler_name += f".{formatter}"
        handler = {
            "class": "logging.StreamHandler",
            "level": level,
            "stream": "ext://sys.stdout",
            "formatter": formatter,
        }
        return handler, handler_name

    def add_handler_stderr(self, filename, formatter, level):
        """Add an handler for stderr."""
        handler_name = f"{filename}.{level.lower()}"
        if formatter in ("django.server", "colorized") and not self.stderr.isatty():
            formatter = None
        elif formatter:
            handler_name += f".{formatter}"
        handler = {
            "class": "logging.StreamHandler",
            "level": level,
            "stream": "ext://sys.stderr",
            "formatter": formatter,
        }
        return handler, handler_name

    @staticmethod
    def get_smart_command_name(module_name, argv, excluded_commands=None):
        """Return a "smart" name for the current command line.

        If it's an interactive Django command (think to "migrate"), returns None
        Otherwise, add the Django command in the name.

        :param module_name:
        :param argv:
        :param excluded_commands:
        :return:
        """
        # command_name = LogConfiguration.resolve_command()
        first_arg = argv[1] if len(argv) >= 2 else None
        if excluded_commands and first_arg in excluded_commands:
            return None
        log_suffix = "%s-%s" % (module_name, first_arg)
        return log_suffix

    @staticmethod
    def resolve_command():
        """Extract the command name in stack traces."""
        f = extract_stack()
        for filename, line_number, name, text in f:
            if filename.endswith("df_config/manage.py") and name in (
                "celery",
                "server",
            ):
                return name
        return None


log_configuration = LogConfiguration()



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <df_config@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################

from django.contrib.staticfiles.management.commands.collectstatic import (
    Command as BaseCommand,
)

from df_config.config.base import merger


# noinspection PyClassHasNoInit
class Command(BaseCommand):
    def handle(self, **options):
        merger.call_method_on_config_values("pre_collectstatic")
        super().handle(**options)
        merger.call_method_on_config_values("post_collectstatic")



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <df_config@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
import base64
import binascii
import logging
from functools import lru_cache

from django.conf import settings
from django.contrib import auth
from django.contrib.auth.middleware import (
    RemoteUserMiddleware as BaseRemoteUserMiddleware,
)
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest
from django.utils.functional import cached_property

logger = logging.getLogger("django.request")


class DFConfigMiddleware(BaseRemoteUserMiddleware):
    """Like :class:`django.contrib.auth.middleware.RemoteUserMiddleware` but:

    * can use any header defined by the setting `DF_REMOTE_USER_HEADER`,
    * handle the HTTP_X_FORWARDED_FOR HTTP header (set the right client IP)
    * handle HTTP basic authentication
    * set response header for Internet Explorer (to use its most recent render engine)
    """

    @lru_cache()
    def get_remoteuser_header(self):
        # avoid cached_property to ease unittests
        header = settings.DF_REMOTE_USER_HEADER
        if header:
            header = header.upper().replace("-", "_")
        return header

    def process_request(self, request: HttpRequest):
        request.remote_username = None

        if settings.USE_X_FORWARDED_FOR and "HTTP_X_FORWARDED_FOR" in request.META:
            request.META["REMOTE_ADDR"] = (
                request.META["HTTP_X_FORWARDED_FOR"].split(",")[0].strip()
            )

        if settings.USE_HTTP_BASIC_AUTH and "HTTP_AUTHORIZATION" in request.META:
            authentication = request.META["HTTP_AUTHORIZATION"]
            authmeth, sep, auth_data = authentication.partition(" ")
            if sep == " " and authmeth.lower() == "basic":
                try:
                    auth_data = base64.b64decode(auth_data.strip()).decode("utf-8")
                except binascii.Error:
                    auth_data = ""
                except UnicodeDecodeError:
                    auth_data = ""
                username, sep, password = auth_data.partition(":")
                if sep == ":":
                    user = auth.authenticate(username=username, password=password)
                    if user:
                        request.user = user
                        auth.login(request, user)
        username = getattr(settings, "DF_FAKE_AUTHENTICATION_USERNAME", None)

        header = self.get_remoteuser_header()
        if header and username and settings.DEBUG:
            remote_addr = request.META.get("REMOTE_ADDR")
            if remote_addr in settings.INTERNAL_IPS:
                request.META[header] = username
            elif remote_addr:
                logger.warning(
                    "Unable to use `settings.DF_FAKE_AUTHENTICATION_USERNAME`. "
                    "You should add %s to the list `settings.INTERNAL_IPS`."
                    % remote_addr
                )

        if header and header in request.META:
            remote_username = request.META.get(header)
            if (
                not remote_username or remote_username == "(null)"
            ):  # special case due to apache2+auth_mod_kerb :-(
                return
            remote_username = self.format_remote_username(remote_username)
            # noinspection PyTypeChecker
            self.remote_user_authentication(request, remote_username)

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def process_response(self, request, response):
        response["X-UA-Compatible"] = "IE=edge,chrome=1"
        return response

    def remote_user_authentication(self, request, username):
        # AuthenticationMiddleware is required so that request.user exists.
        # noinspection PyTypeChecker
        if not hasattr(request, "user"):
            raise ImproperlyConfigured(
                "The Django remote user auth middleware requires the"
                " authentication middleware to be installed.  Edit your"
                " MIDDLEWARE_CLASSES setting to insert"
                " 'django.contrib.auth.middleware.AuthenticationMiddleware'"
                " before the RemoteUserMiddleware class."
            )
        # If the user is already authenticated and that user is the user we are
        # getting passed in the headers, then the correct user is already
        # persisted in the session and we don't need to continue.
        if request.user.is_authenticated:
            cleaned_username = self.clean_username(username, request)
            if request.user.get_username() == cleaned_username:
                request.remote_username = cleaned_username
                return
            else:
                self._remove_invalid_user(request)
        # We are seeing this user for the first time in this session, attempt
        # to authenticate the user.
        user = auth.authenticate(remote_user=username)
        if user:
            # User is valid.  Set request.user and persist user in the session
            # by logging the user in.
            request.user = user
            auth.login(request, user)
            request.remote_username = user.username

    # noinspection PyMethodMayBeStatic
    def format_remote_username(self, remote_username):
        return remote_username.partition("@")[0]



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <df_config@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################


def static_storage(settings_dict):
    if settings_dict["USE_WHITENOISE"] and settings_dict["PIPELINE_ENABLED"]:
        return "df_config.apps.pipeline.PipelineCompressedManifestStaticFilesStorage"
    elif settings_dict["USE_WHITENOISE"]:
        return "whitenoise.storage.CompressedManifestStaticFilesStorage"
    elif settings_dict["PIPELINE_ENABLED"]:
        return "df_config.apps.pipeline.NicerPipelineCachedStorage"
    return "django.contrib.staticfiles.storage.StaticFilesStorage"


static_storage.required_settings = ["PIPELINE_ENABLED", "USE_WHITENOISE"]


def static_storage_setting(settings_dict):
    if settings_dict["USE_WHITENOISE"] and settings_dict["PIPELINE_ENABLED"]:
        backend = "df_config.apps.pipeline.PipelineCompressedManifestStaticFilesStorage"
    elif settings_dict["USE_WHITENOISE"]:
        backend = "whitenoise.storage.CompressedManifestStaticFilesStorage"
    elif settings_dict["PIPELINE_ENABLED"]:
        backend = "df_config.apps.pipeline.NicerPipelineCachedStorage"
    else:
        backend = "django.contrib.staticfiles.storage.StaticFilesStorage"
    options = {
        "location": settings_dict["STATIC_ROOT"],
        "base_url": settings_dict["STATIC_URL"],
    }
    return {"BACKEND": backend, "OPTIONS": options}


static_storage_setting.required_settings = [
    "DEBUG",
    "PIPELINE_ENABLED",
    "USE_WHITENOISE",
    "STATIC_ROOT",
    "STATIC_URL",
]


def media_storage_setting(settings_dict):
    backend = "django.core.files.storage.FileSystemStorage"
    options = {
        "location": settings_dict["MEDIA_ROOT"],
        "base_url": settings_dict["MEDIA_URL"],
    }
    return {"BACKEND": backend, "OPTIONS": options}


media_storage_setting.required_settings = ["MEDIA_ROOT", "MEDIA_URL"]


def pipeline_enabled(settings_dict):
    return settings_dict["USE_PIPELINE"] and not settings_dict["DEBUG"]


pipeline_enabled.required_settings = ["DEBUG", "USE_PIPELINE"]


def static_finder(settings_dict):
    r = [
        "django.contrib.staticfiles.finders.FileSystemFinder",
        "django.contrib.staticfiles.finders.AppDirectoriesFinder",
    ]
    if settings_dict["USE_PIPELINE"]:
        r.append("pipeline.finders.PipelineFinder")
    return r


static_finder.required_settings = ["DEBUG", "USE_PIPELINE"]



# #############################################################################
#  This file is part of df_config                                             #
#                                                                             #
#  Copyright (C) 2020 Matthieu Gallet <df_config@19pouces.net>                #
#  All Rights Reserved                                                        #
#                                                                             #
#  You may use, distribute and modify this code under the                     #
#  terms of the (BSD-like) CeCILL-B license.                                  #
#                                                                             #
#  You should have received a copy of the CeCILL-B license with               #
#  this file. If not, please visit:                                           #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)          #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)        #
#                                                                             #
# #############################################################################
"""__init__ file."""
from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("df_config")
except PackageNotFoundError:
    __version__ = "1.0.0"



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <df_config@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
from urllib.parse import urlencode, urlparse

from django.core.exceptions import ImproperlyConfigured
from django.utils.version import get_complete_version

from df_config.config.url import DatabaseURL
from df_config.utils import is_package_present


def databases(settings_dict):
    """Build a complete DATABASES setting.

    If present, Takes the `DATABASE_URL` environment variable into account
    (used on the Heroku platform).
    """
    default = {
        "ENGINE": DatabaseURL.normalize_engine(settings_dict["DATABASE_ENGINE"]),
        "NAME": settings_dict["DATABASE_NAME"],
        "USER": settings_dict["DATABASE_USER"],
        "OPTIONS": settings_dict["DATABASE_OPTIONS"],
        "PASSWORD": settings_dict["DATABASE_PASSWORD"],
        "HOST": settings_dict["DATABASE_HOST"],
        "PORT": settings_dict["DATABASE_PORT"],
    }
    return {"default": default}


databases.required_settings = [
    "DATABASE_ENGINE",
    "DATABASE_NAME",
    "DATABASE_USER",
    "DATABASE_OPTIONS",
    "DATABASE_PASSWORD",
    "DATABASE_HOST",
    "DATABASE_PORT",
]


class RedisSmartSetting:
    """Handle values required for Redis configuration, as well as Heroku's standard environment variables.
    Can be used as :class:`df_config.config.dynamic_settings.CallableSetting`.
    """

    _config_values = ["PROTOCOL", "HOST", "PORT", "DB", "PASSWORD"]

    def __init__(
        self,
        prefix="",
        env_variable="REDIS_URL",
        fmt="url",
        extra_values=None,
        only_redis: bool = True,
    ):
        """Build Redis connection parameters from a set of settings:
            %(prefix)sPROTOCOL, %(prefix)sHOST, %(prefix)sPORT, %(prefix)sDB, %(prefix)sPASSWORD.

        :param prefix: prefix of all settings
        :param env_variable: if this environment variable is present, override given settings
        :param fmt: output format:
            "url": redis://:password@host:port/db
            "dict": {"host": "host", "password": "password" or None, "port": port, "db": db}
            "channels": expected by django-channels
        :param extra_values: added to the output format
        """
        self.fmt = fmt
        self.prefix = prefix
        self.env_variable = env_variable
        self.only_redis = only_redis
        self.config_values = list(self._config_values)
        if not only_redis:
            self.config_values += ["USERNAME"]
        self.required_settings = [prefix + x for x in self.config_values]
        self.extra_values = extra_values

    def __call__(self, settings_dict):
        values = {x: settings_dict[self.prefix + x] for x in self.config_values}
        values.setdefault("USERNAME")
        values["AUTH"] = ""
        if values["PASSWORD"]:
            values["AUTH"] = "%s:%s@" % (values["USERNAME"] or "", values["PASSWORD"])
        if self.fmt == "url":
            url = "%(PROTOCOL)s://%(AUTH)s%(HOST)s:%(PORT)s/%(DB)s" % values
            if self.extra_values:
                url += "?" + urlencode(self.extra_values)
            return url
        elif self.fmt == "channels":
            url = "%(PROTOCOL)s://%(AUTH)s%(HOST)s:%(PORT)s/%(DB)s" % values
            config = {
                "address": url,
                "password": values["PASSWORD"] or None,
                "db": int(values["DB"] or 0),
            }
            if self.extra_values:
                config.update(self.extra_values)
            # noinspection PyUnresolvedReferences
            result = {
                "BACKEND": "channels_redis.core.RedisChannelLayer",
                "CONFIG": {
                    "hosts": [config],
                    "capacity": 5000,
                    "expiry": 10,
                },
            }
            return result
        elif self.fmt == "dict":
            result = {
                "host": values["HOST"] or "localhost",
                "port": int(values["PORT"] or 6379),
                "db": int(values["DB"] or 0),
                "password": values["PASSWORD"] or None,
            }
            if self.extra_values:
                result.update(self.extra_values)
            return result
        raise ValueError("Unknown RedisSmartSetting format '%s'" % self.fmt)

    def __repr__(self):
        p = self.prefix
        if self.prefix.endswith("REDIS_"):
            p = self.prefix[:-6]
        return "%s.%sredis_%s" % (self.__module__, p.lower(), self.fmt)


cache_redis_url = RedisSmartSetting(prefix="CACHE_", fmt="url")
celery_broker_url = RedisSmartSetting(prefix="CELERY_", fmt="url", only_redis=False)
celery_result_url = RedisSmartSetting(
    prefix="CELERY_RESULT_", fmt="url", only_redis=False
)
session_redis_dict = RedisSmartSetting(
    prefix="SESSION_REDIS_", fmt="dict", extra_values={"prefix": "session"}
)
websocket_redis_dict = RedisSmartSetting(prefix="WEBSOCKET_REDIS_", fmt="dict")
websocket_redis_channels = RedisSmartSetting(prefix="WEBSOCKET_REDIS_", fmt="channels")


def cache_setting(settings_dict):
    """Automatically compute cache settings:
      * if debug mode is set, then caching is disabled
      * if django_redis is available, then Redis is used for caching
      * else memory is used

    :param settings_dict:
    :return:
    """
    parsed_url = urlparse(settings_dict["CACHE_URL"])
    django_version = get_complete_version()
    default = {
        "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
        "LOCATION": "unique-snowflake",
    }
    if settings_dict["DEBUG"]:
        pass
    elif django_version >= (4, 0) and parsed_url.scheme in ("redis", "rediss"):
        # noinspection PyUnresolvedReferences
        default = {
            "BACKEND": "django.core.cache.backends.redis.RedisCache",
            "LOCATION": "{CACHE_URL}",
        }
    elif parsed_url.scheme in ("redis", "rediss"):
        if is_package_present("django_redis"):
            # noinspection PyUnresolvedReferences
            default = {
                "BACKEND": "django_redis.cache.RedisCache",
                "LOCATION": "{CACHE_URL}",
                "OPTIONS": {"CLIENT_CLASS": "django_redis.client.DefaultClient"},
            }
    elif parsed_url.scheme == "memcache":
        if django_version >= (3, 2) and is_package_present("pymemcache"):
            backend = "django.core.cache.backends.memcached.PyMemcacheCache"
        elif django_version >= (3, 2) and is_package_present("pylibmc"):
            backend = "django.core.cache.backends.memcached.PyLibMCCache"
        elif is_package_present("memcache"):
            backend = "django.core.cache.backends.memcached.MemcachedCache"
        else:
            raise ImproperlyConfigured(
                "Please install 'pylibmc' package before using memcache engine."
            )
        location = "%s:%s" % (
            parsed_url.hostname or "localhost",
            parsed_url.port or 11211,
        )
        default = {
            "BACKEND": backend,
            "LOCATION": location,
        }
    return {"default": default}


cache_setting.required_settings = ["DEBUG", "CACHE_URL"]



from df_config.utils import is_package_present

DEBUG_TOOLBAR_PANELS = [
    "debug_toolbar.panels.versions.VersionsPanel",
    "debug_toolbar.panels.timer.TimerPanel",
    "debug_toolbar.panels.settings.SettingsPanel",
    "debug_toolbar.panels.profiling.ProfilingPanel",
    "debug_toolbar.panels.headers.HeadersPanel",
    "debug_toolbar.panels.request.RequestPanel",
    "debug_toolbar.panels.sql.SQLPanel",
    "debug_toolbar.panels.templates.TemplatesPanel",
    "debug_toolbar.panels.staticfiles.StaticFilesPanel",
    "debug_toolbar.panels.cache.CachePanel",
    "debug_toolbar.panels.signals.SignalsPanel",
    "debug_toolbar.panels.redirects.RedirectsPanel",
]


def guess_djt_panels(settings_dict):
    if not settings_dict["DEBUG"] or not settings_dict["USE_DEBUG_TOOLBAR"]:
        return []
    panels = list(DEBUG_TOOLBAR_PANELS)
    if is_package_present("djt_og"):
        panels.insert(2, "djt_og.panel.OpenGraphPanel")
    if is_package_present("djt_csp"):
        panels.insert(2, "djt_csp.panel.SecurityPanel")
    if is_package_present("djt_nvu"):
        panels.insert(2, "djt_nvu.panel.W3ValidatorPanel")
    return panels


guess_djt_panels.required_settings = ["DEBUG", "USE_DEBUG_TOOLBAR"]



from queue import Queue
from typing import Optional, Tuple

from logging_loki import LokiQueueHandler, emitter

emitter.LokiEmitter.level_tag = "level"


class LokiHandler(LokiQueueHandler):
    def __init__(
        self, url: Optional[str] = None, auth: Optional[Tuple[str, str]] = None
    ):
        from django.conf import settings

        tags = {
            "application": settings.SERVER_NAME,
            "command": settings.CURRENT_COMMAND_NAME,
            "hostname": settings.HOSTNAME,
            "log_source": "django",
        }
        super().__init__(Queue(-1), url=url, tags=tags, auth=auth, version="1")



# ##############################################################################
#  This file is part of df_config                                              #
#                                                                              #
#  Copyright (C) 2020 Matthieu Gallet <df_config@19pouces.net>                    #
#  All Rights Reserved                                                         #
#                                                                              #
#  You may use, distribute and modify this code under the                      #
#  terms of the (BSD-like) CeCILL-B license.                                   #
#                                                                              #
#  You should have received a copy of the CeCILL-B license with                #
#  this file. If not, please visit:                                            #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)           #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)         #
#                                                                              #
# ##############################################################################
import hashlib
import importlib.util
import os
import shlex
from collections import OrderedDict
from configparser import ConfigParser
from importlib import import_module
from io import StringIO
from typing import Any, Iterable, Tuple

from df_config.config.fields import ConfigField
from df_config.config.fields_providers import import_attribute


class ConfigProvider:
    """Base class of config provider."""

    name = None

    def has_value(self, config_field: ConfigField) -> bool:
        """Return True if a config_field is present in the file"""
        raise NotImplementedError

    def set_value(self, config_field: ConfigField, include_doc=False):
        """Get the value of the config_field and set its internal value"""
        raise NotImplementedError

    def get_value(self, config_field: ConfigField):
        """Get the internal value if the config field is present in its internal values.
        Otherwise returns the current value of the config field.
        """
        raise NotImplementedError

    def get_extra_settings(self) -> Iterable[Tuple[str, Any]]:
        """Return all settings internally defined.

        :return: an iterable of (setting_name, value).
        """
        raise NotImplementedError

    def is_valid(self) -> bool:
        """Return `True` if the provider is valid (for example, the corresponding file is missing)."""
        raise NotImplementedError

    def to_str(self) -> str:
        """Convert all its internal values to a string"""
        raise NotImplementedError


class EnvironmentConfigProvider(ConfigProvider):
    name = "Environment"

    def __init__(self, prefix, mapping: str = None):

        self.prefix = prefix
        self.exports = ""
        self.exported_values = set()

    def __str__(self):
        count = len(self.exported_values)
        if count <= 1:
            return f"Shell environment ({count} variable)"
        return f"Shell environment ({count} variables)"

    def has_value(self, config_field: ConfigField):
        key = self.get_key(config_field)
        if key is None:
            return False
        return key in os.environ

    def get_key(self, config_field):
        if config_field.environ_name is config_field.AUTO:
            key = f"{self.prefix}{config_field.setting_name}"
        else:
            key = config_field.environ_name
        return key

    def set_value(self, config_field, include_doc=False):
        key = self.get_key(config_field)
        if key is None:
            return
        value = config_field.to_str(config_field.value)
        self.exports += "%s=%s\n" % (shlex.quote(key), shlex.quote(value))

    def get_value(self, config_field):
        key = self.get_key(config_field)
        if key in os.environ:
            self.exported_values.add(key)
            return config_field.from_str(os.environ[key])
        return config_field.value

    def get_extra_settings(self):
        return []

    def is_valid(self):
        return True

    def to_str(self):
        return self.exports


class IniConfigProvider(ConfigProvider):
    """Read a config file using the .ini syntax."""

    name = ".ini file"

    def __init__(self, config_file=None):
        self.parser = ConfigParser()
        self.config_file = config_file
        if config_file:
            self.parser.read([config_file])

    def __str__(self):
        return self.config_file

    @staticmethod
    def __get_info(config_field: ConfigField):
        if config_field.name is None:
            return None, None
        section, sep, option = config_field.name.partition(".")
        return section, option

    def set_value(self, config_field: ConfigField, include_doc: bool = False):
        """update the internal config file"""
        section, option = self.__get_info(config_field)
        if section is None:
            return
        if not self.parser.has_section(section):
            self.parser.add_section(section)
        to_str = config_field.to_str(config_field.value)
        if include_doc and config_field.__doc__:
            for line in config_field.__doc__.splitlines():
                to_str += " \n# %s" % line
        self.parser.set(section, option, to_str)

    def get_value(self, config_field: ConfigField):
        """get option from the config file"""
        section, option = self.__get_info(config_field)
        if section is None:
            return None
        if self.parser.has_option(section=section, option=option):
            str_value = self.parser.get(section=section, option=option)
            return config_field.from_str(str_value)
        return config_field.value

    def has_value(self, config_field: ConfigField):
        """return `True` if the option is defined in the config file"""
        section, option = self.__get_info(config_field)
        if section is None:
            return False
        return self.parser.has_option(section=section, option=option)

    def get_extra_settings(self):
        """No extra setting can be defined in a config file"""
        return []

    def is_valid(self):
        """Return `True` if the config file exists"""
        return os.path.isfile(self.config_file)

    def to_str(self):
        """Display the config file"""
        fd = StringIO()
        self.parser.write(fd)
        return fd.getvalue()


class PythonModuleProvider(ConfigProvider):
    """Load a Python module from its dotted name"""

    name = "Python module"

    def __init__(self, module_name=None):
        self.module_name = module_name
        self.module = None
        self.values = OrderedDict()
        if module_name is not None:
            try:
                self.module = import_module(module_name, package=None)
            except AttributeError:
                pass
            except ImportError:
                pass

    def __str__(self):
        return self.module_name

    def set_value(self, config_field, include_doc=False):
        """Set the value of the config field in an internal dict"""
        self.values[config_field.setting_name] = config_field.value

    def get_value(self, config_field):
        """Get the value of a variable defined in the Python module."""
        if self.module is None or not hasattr(self.module, config_field.setting_name):
            return config_field.value
        return getattr(self.module, config_field.setting_name)

    def has_value(self, config_field):
        """Return `True` if the corresponding variable is defined in the module"""
        return self.module is not None and hasattr(
            self.module, config_field.setting_name
        )

    def get_extra_settings(self):
        """Return all values that look like a Django setting (i.e. uppercase variables)"""
        if self.module is not None:
            for key, value in sorted(self.module.__dict__.items()):
                if key.upper() != key or key == "_":
                    continue
                yield key, value

    def is_valid(self):
        """Return `True` if the module can be imported"""
        return bool(self.module)

    def to_str(self):
        """Display values as if set in a Python module"""
        fd = StringIO()
        for k, v in sorted(self.values.items()):
            fd.write("%s = %r\n" % (k, v))
        return fd.getvalue()


class PythonFileProvider(PythonModuleProvider):
    """Load a Python module from its absolute path"""

    name = "Python file"

    def __init__(self, module_filename):
        self.module_filename = module_filename
        super().__init__()
        if not os.path.isfile(module_filename):
            return
        md5 = hashlib.md5(module_filename.encode("utf-8")).hexdigest()  # nosec  # nosec
        module_name = "df_config.__private" + md5
        spec = importlib.util.spec_from_file_location(module_name, module_filename)
        module_ = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module_)
        self.module = module_

    def __str__(self):
        return self.module_filename


class DictProvider(ConfigProvider):
    """Use a plain Python dict as a setting provider"""

    name = "dict"

    def __init__(self, values, name="default values"):
        self.name = name
        self.values = values

    def get_extra_settings(self):
        """Return all uppercase keys of the internal dict as valid filenames"""
        for k, v in self.values.items():
            if k == k.upper():
                yield k, v

    def set_value(self, config_field, include_doc=False):
        """modify the internal dict for storing the value"""
        self.values[config_field.setting_name] = config_field.value

    def get_value(self, config_field):
        """get a value from the internal dict if present"""
        return self.values.get(config_field.setting_name, config_field.value)

    def has_value(self, config_field):
        """check if the value is present in the internal dict"""
        return config_field.setting_name in self.values

    def __str__(self):
        return self.name

    def is_valid(self):
        """Return always `True`"""
        return True

    def to_str(self):
        """Display the internal dict"""
        return "%r" % dict(self.values)
